﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Emgu.CV;
using Emgu.CV.Structure;
using System.Drawing;

using bbiwarg.Detectors.Fingers;
using bbiwarg.Utility;

namespace bbiwarg.Images
{
    class PalmImage
    {
        DepthImage depthImage;
        EdgeImage edgeImage;

        private Image<Gray, Byte> handImage;
        private Image<Gray, Byte> pointingHandMask;
        private Image<Gray, Byte> outputImage;

        private List<Finger> fingers;
        
        private int width, height;
        
        public PalmImage(DepthImage depthImage, EdgeImage edgeImage, FingerDetector fingerDetector)
        {
            this.depthImage = depthImage;
            this.edgeImage = edgeImage;

            handImage = depthImage.getImage().Convert<Byte>(delegate(short s) { return (s == depthImage.getMaxDepth()) ? (byte) 0 : (byte) 1; });
            width = depthImage.getWidth();
            height = depthImage.getHeight();

            fingers = getFingersWithoutThumb(fingerDetector);
            buildPointingHandMask();
            handImage = handImage.And(pointingHandMask);

            findPalmContour();
        }

        private List<Finger> getFingersWithoutThumb(FingerDetector fingerDetector)
        {
            List<Finger> detectedFingers = fingerDetector.Fingers;
            Finger leftMost = null;
            float minX = float.MaxValue;
            foreach (Finger f in detectedFingers)
            {
                float midX = ((f.Hand + f.Tip) / 2).X;
                if (midX < minX)
                {
                    minX = midX;
                    leftMost = f;
                }
            }

            List<Finger> result = new List<Finger>();
            foreach (Finger f in detectedFingers)
            {
                if (f != leftMost)
                    result.Add(f);
            }
            return result;
        }

        private void fillFingerSlices(Image<Gray, Byte> image, byte val)
        {
            foreach (Finger f in fingers)
            {
                foreach (FingerSlice s in f.SliceTrail.getSlices())
                {
                    image.Draw(new LineSegment2DF(s.Start, s.End), new Gray(val), 1);
                }            
            }
        }

        private Finger getLongestFinger()
        {
            float maxLength = 0;
            Finger longest = null;
            foreach (Finger f in fingers)
            {
                
                if (f.Line.Length > maxLength)
                {
                    maxLength = f.Line.Length;
                    longest = f;
                }
            }
            return longest;
        }

        private Point getPointInPointingHand()
        {
            Finger finger = getLongestFinger();
            if (finger == null)
                return new Point(0, 0);
            Vector2D direction = (finger.Hand - finger.Tip).normalize();
            Vector2D pos = finger.Hand + 20 * direction;
            return new Point(Math.Min(width - 1, Math.Max(0, (int)pos.X)), Math.Min(height - 1, Math.Max(0, (int)pos.Y)));
        }

        private void buildPointingHandMask()
        {
            pointingHandMask = new Image<Gray, byte>(width, height, new Gray(0));
            fillFingerSlices(pointingHandMask, 1);
            pointingHandMask = pointingHandMask.Dilate(1);
            pointingHandMask = pointingHandMask.Or(edgeImage.getImage().Convert<Byte>(delegate(byte b) { return (b == 0) ? (byte)0 : (byte)1; }));
            pointingHandMask = pointingHandMask.Dilate(1);

            MCvConnectedComp tmp = new MCvConnectedComp();
            CvInvoke.cvFloodFill(pointingHandMask.Ptr, getPointInPointingHand(), new MCvScalar(2), new MCvScalar(0), new MCvScalar(0), out tmp, 0, IntPtr.Zero);
            pointingHandMask = pointingHandMask.Convert<Byte>(delegate(byte b) { return (b == 2) ? (byte)0 : (byte)1; });
            pointingHandMask = pointingHandMask.Erode(1);
            fillFingerSlices(pointingHandMask, 0);
            pointingHandMask = pointingHandMask.Erode(2);
        }

        private void findPalmContour()
        {
            Contour<Point> contour = handImage.FindContours(Emgu.CV.CvEnum.CHAIN_APPROX_METHOD.CV_CHAIN_APPROX_SIMPLE, Emgu.CV.CvEnum.RETR_TYPE.CV_RETR_EXTERNAL);

            Contour<Point> maxContour = contour;
            double maxPerimeter = 0;
            while (contour != null)
            {
                if (contour.Perimeter > maxPerimeter)
                {
                    maxPerimeter = contour.Perimeter;
                    maxContour = contour;
                }
                contour = contour.HNext;
            }

            if (maxContour != null)
            {
                Contour<Point> contourPoly = maxContour; //.ApproxPoly(0);
                Seq<Point> hull = contourPoly.GetConvexHull(Emgu.CV.CvEnum.ORIENTATION.CV_CLOCKWISE);
                Seq<MCvConvexityDefect> defects = contourPoly.GetConvexityDefacts(new MemStorage(), Emgu.CV.CvEnum.ORIENTATION.CV_CLOCKWISE);

                outputImage = new Image<Gray, byte>(width, height, new Gray(0));
                outputImage.Draw(contourPoly, new Gray(1), 1);
                outputImage.Draw(hull, new Gray(1), 1);

                foreach (MCvConvexityDefect defect in defects)
                {
                    Vector2D center = (new Vector2D(defect.StartPoint) + new Vector2D(defect.EndPoint)) / 2;

                    float minFingerDist = float.MaxValue;
                    foreach (Finger f in fingers)
                    {
                        float dist = f.Tip.getDistanceTo(new Vector2D(defect.DepthPoint));
                        if (dist < minFingerDist)
                            minFingerDist = dist;
                    }

                    if (minFingerDist > 20)
                    {
                        //Console.WriteLine(minFingerDist);
                        outputImage.Draw(new LineSegment2DF(defect.DepthPoint, center), new Gray(1), 1);
                    }
                }
                //Console.WriteLine();
            }
        }

        public bool belongsToPalm(int x, int y)
        {
            if (outputImage != null)
                return outputImage.Data[y, x, 0] == 1;
            return false;
        }
    }
}
